#!/usr/bin/python

import time
import struct
import logging
import sys

slow_xor = False
imported = False

BLOCK_SIZE = 16384


def run_imports():
    global imported, slow_xor, salsa20, numpy
    if not imported:
        imported = True
        try:
            import numpy
        except ImportError:
            logging.error('can not import numpy, using SLOW XOR')
            logging.error('please install numpy if you use salsa20')
            slow_xor = True
        try:
            import salsa20
        except ImportError:
            logging.error('you have to install salsa20 before you use salsa20')
            sys.exit(1)


def numpy_xor(a, b):
    if slow_xor:
        return py_xor_str(a, b)
    dtype = numpy.byte
    if len(a) % 4 == 0:
        dtype = numpy.uint32
    elif len(a) % 2 == 0:
        dtype = numpy.uint16

    ab = numpy.frombuffer(a, dtype=dtype)
    bb = numpy.frombuffer(b, dtype=dtype)
    c = numpy.bitwise_xor(ab, bb)
    r = c.tostring()
    return r


def py_xor_str(a, b):
    c = []
    for i in xrange(0, len(a)):
        c.append(chr(ord(a[i]) ^ ord(b[i])))
    return ''.join(c)


class Salsa20Cipher(object):
    """a salsa20 CTR implemetation, provides m2crypto like cipher API"""

    def __init__(self, alg, key, iv, op, key_as_bytes=0, d=None, salt=None,
                 i=1, padding=1):
        run_imports()
        if alg != 'salsa20-ctr':
            raise Exception('unknown algorithm')
        self._key = key
        self._nonce = struct.unpack('<Q', iv)[0]
        self._pos = 0
        self._next_stream()

    def _next_stream(self):
        self._nonce &= 0xFFFFFFFFFFFFFFFF
        self._stream = salsa20.Salsa20_keystream(BLOCK_SIZE,
                                                 struct.pack('<Q', self._nonce),
                                                 self._key)
        self._nonce += 1

    def update(self, data):
        results = []
        while True:
            remain = BLOCK_SIZE - self._pos
            cur_data = data[:remain]
            cur_data_len = len(cur_data)
            cur_stream = self._stream[self._pos:self._pos + cur_data_len]
            self._pos = self._pos + cur_data_len
            data = data[remain:]

            results.append(numpy_xor(cur_data, cur_stream))

            if self._pos >= BLOCK_SIZE:
                self._next_stream()
                self._pos = 0
            if not data:
                break
        return ''.join(results)


def test():
    from os import urandom
    import random

    rounds = 1 * 1024
    plain = urandom(BLOCK_SIZE * rounds)
    import M2Crypto.EVP
    # cipher = M2Crypto.EVP.Cipher('aes_128_cfb', 'k' * 32, 'i' * 16, 1,
    #                key_as_bytes=0, d='md5', salt=None, i=1,
    #                padding=1)
    # decipher = M2Crypto.EVP.Cipher('aes_128_cfb', 'k' * 32, 'i' * 16, 0,
    #                key_as_bytes=0, d='md5', salt=None, i=1,
    #                padding=1)

    cipher = Salsa20Cipher('salsa20-ctr', 'k' * 32, 'i' * 8, 1)
    decipher = Salsa20Cipher('salsa20-ctr', 'k' * 32, 'i' * 8, 1)
    results = []
    pos = 0
    print 'salsa20 test start'
    start = time.time()
    while pos < len(plain):
        l = random.randint(100, 32768)
        c = cipher.update(plain[pos:pos + l])
        results.append(c)
        pos += l
    pos = 0
    c = ''.join(results)
    results = []
    while pos < len(plain):
        l = random.randint(100, 32768)
        results.append(decipher.update(c[pos:pos + l]))
        pos += l
    end = time.time()
    print 'speed: %d bytes/s' % (BLOCK_SIZE * rounds / (end - start))
    assert ''.join(results) == plain


if __name__ == '__main__':
    test()


#!/usr/bin/env python

# Copyright (c) 2014 apporc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import pygtk
pygtk.require('2.0')
import gtk
import vte
import subprocess
import gettext
from pkg_resources import resource_filename
from shadowsocks_gtk.encrypt import method_supported
from shadowsocks_gtk.utils import get_config, save_config


DEFAULT_LOGLEVEL = 'info'
WINDOW_HEIGHT = 260
WINDOW_WIDTH = 450
VTE_HEIGHT = 200
LINE_HEIGHT = 25
LINE_HEADER_WIDTH = 120
ADDRESS_WIDTH = 180
PORT_WIDTH = 50
LOGO_FILE = 'shadowsocks.png'
DEFAULT_SERVER = ['209.141.36.62', '8348', '$#HAL9000!', 'aes-256-cfb']
DEFAULT_LOCAL = ['127.0.0.1', '1080']
DEFAULT_TIMEOUT = 600


gettext.bindtextdomain('shadowsocks_gtk', resource_filename(__name__, 'locale'))
gettext.textdomain('shadowsocks_gtk')
_ = gettext.gettext


class ShadowSocks(object):
    command = ['/usr/bin/env', 'python', 'local.py']

    def __init__(self, window=None):
        self.supported_methods = method_supported.keys()
        self.config = get_config()
        server_ip = self.config.get('server', None)
        server_port = self.config.get('server_port', None)
        self.servers = self.config.get('servers', [DEFAULT_SERVER])
        if server_ip and server_port:
            self.server_ip = str(server_ip)
            self.server_port = str(server_port)
            self.password = self.config.get('password')
            method = self.config.get('method')
        else:
            self.server_ip = DEFAULT_SERVER[0]
            self.server_port = DEFAULT_SERVER[1]
            self.password = DEFAULT_SERVER[2]
            method = DEFAULT_SERVER[3]

        self.method = self.supported_methods.index(method) if (method in self.supported_methods) else 0
        self.local_ip = self.config.get('local_address', DEFAULT_LOCAL[0])
        self.local_port = str(self.config.get('local_port', DEFAULT_LOCAL[1]))
        self.timeout = str(self.config.get('timeout', DEFAULT_TIMEOUT))
        self.status = 'disconnected'
        self.childpid = None
        self.childexited = True

        self.window = window
        self.labels = {}
        self.entrys = {}
        self.buttons = {}
        if not self.window:
            self.create_window()
        self.create_trayicon()
        self.create_menu()
        self.fill_window()

    def create_window(self):
        # create a new window
        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.window.set_title('ShadowSocks')
        self.window.set_position(gtk.WIN_POS_CENTER)
        self.window.set_border_width(10)
        self.window.connect('window-state-event', self.state_event)
        self.window.connect("destroy", self.destroy)
        self.window.set_size_request(WINDOW_WIDTH, WINDOW_HEIGHT)
        logo = resource_filename(__name__, LOGO_FILE)
        self.window.set_icon_from_file(logo)
        self.window.show()

    def create_trayicon(self):
        self.trayicon = gtk.StatusIcon()
        logo = resource_filename(__name__, LOGO_FILE)
        self.trayicon.set_from_file(logo)
        self.trayicon.connect('popup-menu', self.show_menu)
        self.trayicon.connect('activate', self.toggle_showhide_item)
        self.trayicon.set_tooltip('ShadowSocks')
        self.trayicon.set_visible(True)

    def show_menu(self, statusicon, button, activation_time):
        self.menu.popup(None, None, gtk.status_icon_position_menu, button, activation_time, self.trayicon)

    def create_menu(self):
        self.menu = gtk.Menu()
        self.showhide_item = gtk.MenuItem(_('Hide'))
        self.showhide_item.connect('activate', self.toggle_showhide_item)
        self.connect_item = gtk.MenuItem(_('Connect'))
        self.connect_item.connect('activate', self.toggle_connect)
        self.quit_item = gtk.MenuItem(_('Quit'))
        self.quit_item.connect('activate', self.destroy)
        self.menu.append(self.connect_item)
        self.menu.append(self.showhide_item)
        self.menu.append(self.quit_item)
        self.menu.show_all()

    def add_line(self, labelname='Default Label :', elements=None):
        #create HBox
        hbox = gtk.HBox(False, 0)
        self.vbox.pack_start(hbox, False, False, 0)
        #create label header
        label = gtk.Label(_(labelname) + ' :')
        label.set_size_request(LINE_HEADER_WIDTH, LINE_HEIGHT)
        label.set_properties(xalign=1)
        hbox.pack_start(label, False, False, 10)
        label.show()
        label_name = labelname.lower()
        self.labels[label_name] = label

        if elements:
            for e in elements:
                hbox.pack_start(e[0], e[1], e[1], e[2])
        return hbox

    def check_port(self, entry):
        port = entry.get_text()
        if not port.isdigit():
            entry.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("red"))
        else:
            entry.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("black"))

    def update_server(self, box):
        t = box.get_active_text()
        if ':' in t:
            t = t.split(':')
            box.child.set_text(t[0])
            self.entrys['server_port'].set_text(t[1])

    def fill_window(self):
        self.vbox = gtk.VBox(False, 10)
        self.window.add(self.vbox)
        #Server Line
        liststore = gtk.ListStore(str)
        self.entrys['server_ip'] = gtk.ComboBoxEntry(liststore, 0)
        for server in self.servers:
            self.entrys['server_ip'].append_text(server[0] + ':' + server[1])
        self.entrys['server_ip'].child.set_text(self.server_ip)
        self.entrys['server_ip'].connect('changed', self.update_server)
        self.entrys['server_ip'].set_size_request(ADDRESS_WIDTH, LINE_HEIGHT)
        self.labels['server_colon'] = gtk.Label(':')
        self.labels['server_colon'].set_properties(xalign=0.5)
        self.entrys['server_port'] = gtk.Entry()
        self.entrys['server_port'].set_size_request(PORT_WIDTH, LINE_HEIGHT)
        self.entrys['server_port'].set_text(self.server_port)
        elements = [(self.entrys['server_ip'], True, 0),
                    (self.labels['server_colon'], False, 0),
                    (self.entrys['server_port'], False, 0)]
        self.server_hbox = self.add_line('Server Address', elements)
        self.entrys['server_port'].connect('changed', self.check_port)

        #Local Line
        self.entrys['local_ip'] = gtk.Entry()
        self.entrys['local_ip'].set_size_request(ADDRESS_WIDTH, LINE_HEIGHT)
        self.entrys['local_ip'].set_text(self.local_ip)
        self.labels['local_colon'] = gtk.Label(':')
        self.labels['local_colon'].set_properties(xalign=0.5)
        self.entrys['local_port'] = gtk.Entry()
        self.entrys['local_port'].set_size_request(PORT_WIDTH, LINE_HEIGHT)
        self.entrys['local_port'].set_text(self.local_port)
        elements = [(self.entrys['local_ip'], True, 0),
                    (self.labels['local_colon'], False, 0),
                    (self.entrys['local_port'], False, 0)]
        self.local_hbox = self.add_line('Local Listening', elements)

        #Password Line
        self.entrys['password'] = gtk.Entry()
        self.entrys['password'].set_visibility(False)
        if self.password:
            self.entrys['password'].set_text(self.password)
        elements = [(self.entrys['password'], True, 0)]
        self.pass_hbox = self.add_line('Password', elements)

        #Encryption Method Line
        self.entrys['encrypt_method'] = gtk.combo_box_new_text()
        for method in self.supported_methods:
            self.entrys['encrypt_method'].append_text(method)
        self.entrys['encrypt_method'].set_active(self.method)
        elements = [(self.entrys['encrypt_method'], True, 0)]
        self.encrypt_hbox = self.add_line('Encryption Method', elements)

        #Timeout Line
        self.entrys['timeout'] = gtk.Entry()
        self.entrys['timeout'].set_text(self.timeout)
        elements = [(self.entrys['timeout'], True, 0)]
        self.timeout_hbox = self.add_line('Timeout in Second', elements)

        #Status Line
        self.labels['current_status'] = gtk.Label(_(self.status))
        self.labels['current_status'].set_properties(xalign=0)
        self.labels['current_status'].show()
        self.buttons['detail'] = gtk.Button(_('show detail'))
        self.buttons['detail'].set_size_request(80, LINE_HEIGHT)
        self.buttons['detail'].connect("clicked", self.toggle_detail_button, None)
        elements = [(self.labels['current_status'], True, 0),
                    (self.buttons['detail'], False, 0)]
        self.status_hbox = self.add_line('Current Status', elements)

        #Add buttonbox
        self.add_buttonbox()

        #Terminal Window
        self.terminal = vte.Terminal()
        self.terminal.set_size_request(380, 180)
        self.vbox.pack_start(self.terminal, True, True, 0)

        #Show them
        self.server_hbox.show_all()
        self.local_hbox.show_all()
        self.pass_hbox.show_all()
        self.encrypt_hbox.show_all()
        self.timeout_hbox.show_all()
        self.status_hbox.show()
        self.buttonbox.show_all()
        self.vbox.show()

    def add_buttonbox(self):
        #HButtonBox
        self.buttonbox = gtk.HButtonBox()
        self.buttonbox.set_layout(gtk.BUTTONBOX_END)
        self.buttons['connect'] = gtk.Button(_('Connect'))
        self.buttons['close'] = gtk.Button(stock=gtk.STOCK_CLOSE)
        self.buttonbox.add(self.buttons['connect'])
        self.buttonbox.add(self.buttons['close'])
        self.buttons['close'].set_size_request(60, LINE_HEIGHT)
        self.buttons['close'].connect_object("clicked", self.destroy, self.window)
        self.buttons['connect'].set_size_request(60, LINE_HEIGHT)
        self.buttons['connect'].connect("clicked", self.toggle_connect)
        self.vbox.pack_start(self.buttonbox, False, False, 0)

    def show_vte(self):
        self.window.resize(WINDOW_WIDTH, WINDOW_HEIGHT + VTE_HEIGHT)
        self.terminal.show()

    def hide_vte(self):
        self.window.resize(WINDOW_WIDTH, WINDOW_HEIGHT)
        self.terminal.hide()

    def on_child_exited(self, term):
        self.childexited = True
        self.update_status(False)
        self.hide_vte()
        self.hide_detail_button()

    def update_status(self, connect):
        if connect:
            self.status = 'connected'
            self.buttons['connect'].set_label(_('Disconnect'))
            self.connect_item.set_label(_('Disconnect'))
            self.labels['current_status'].modify_fg(gtk.STATE_NORMAL, gtk.gdk.color_parse("green"))
            self.labels['current_status'].set_text('%s %s' %
                                                   (_(self.status + ' to'), self.entrys['server_ip'].child.get_text()))
        else:
            self.status = 'disconnected'
            self.buttons['connect'].set_label(_('Connect'))
            self.connect_item.set_label(_('Connect'))
            self.labels['current_status'].modify_fg(gtk.STATE_NORMAL, gtk.gdk.color_parse("black"))
            self.labels['current_status'].set_text(_(self.status))

    def toggle_showhide_item(self, widget=None):
        if self.showhide_item.get_label() == _('Show'):
            self.window.deiconify()
            self.window.show()
            self.showhide_item.set_label(_('Hide'))
        else:
            self.window.iconify()

    def save(self):
        self.server_ip = self.entrys['server_ip'].child.get_text()
        self.server_port = self.entrys['server_port'].get_text()
        self.local_ip = self.entrys['local_ip'].get_text()
        self.local_port = self.entrys['local_port'].get_text()
        self.password = self.entrys['password'].get_text()
        self.timeout = self.entrys['timeout'].get_text()
        self.method = self.entrys['encrypt_method'].get_active()
        method = self.supported_methods[self.method]
        server = [self.server_ip, self.server_port, self.password, method]
        if server not in self.servers:
            self.servers.append(server)

        self.config['servers'] = self.servers
        self.config['server'] = self.server_ip
        self.config['server_port'] = int(self.server_port)
        self.config['local_address'] = self.local_ip
        self.config['local_port'] = int(self.local_port)
        self.config['password'] = self.password
        self.config['timeout'] = int(self.timeout)
        self.config['level'] = DEFAULT_LOGLEVEL
        self.config['method'] = method
        save_config(self.config)

    def toggle_detail_button(self, widget, data=None):
        if self.buttons['detail'].get_label() == _('show detail'):
            self.buttons['detail'].set_label(_('hide detail'))
            self.show_vte()
        else:
            self.buttons['detail'].set_label(_('show detail'))
            self.hide_vte()

    def show_detail_button(self):
        if self.status == 'connected':
            self.buttons['detail'].set_label(_('show detail'))
        else:
            self.buttons['detail'].set_label(_('hide detail'))
        self.buttons['detail'].show()

    def hide_detail_button(self):
        self.buttons['detail'].hide()

    def run(self):
        if self.childexited:
            self.childpid = self.terminal.fork_command(self.command[0], self.command, os.getcwd())
            if self.childpid > 0:
                self.childexited = False
                self.terminal.connect('child-exited', self.on_child_exited)
            else:
                self.childexited = False

    def stop(self):
        if not self.childexited:
            code0 = subprocess.call(['kill', str(self.childpid)])
            if code0 != 0:
                subprocess.call(['kill', '-9', str(self.childpid)])
            self.childexited = True

    def toggle_connect(self, widget=None):
        if self.status == 'disconnected':
            self.save()
            self.run()
            self.update_status(True)
            self.show_detail_button()
        else:
            self.stop()

    def destroy(self, widget, data=None):
        self.stop()
        gtk.main_quit()

    def state_event(self, window, event):
        if event.changed_mask & gtk.gdk.WINDOW_STATE_ICONIFIED\
                and event.new_window_state & gtk.gdk.WINDOW_STATE_ICONIFIED:
            self.showhide_item.set_label(_('Show'))
            self.window.hide()

    def main(self):
        gtk.main()


def main():
    path = os.path.abspath(__file__)
    if os.path.islink(path):
        path = getattr(os, 'readlink', lambda x: x)(path)
    os.chdir(os.path.dirname(os.path.abspath(path)))

    shadowsocks = ShadowSocks()
    shadowsocks.main()


if __name__ == "__main__":
    main()



# Copyright (c) 2014 apporc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import logging
from twisted.python import log
import json

CONFIG_PATH = os.environ['HOME'] + '/.config/' + 'shadowsocks-gtk/'
CONFIG_FILE = 'config.json'


def get_config():
    path = CONFIG_PATH + CONFIG_FILE
    if os.path.exists(path):
        with open(path, 'rb') as f:
            options = json.load(f)
        return options
    else:
        return {}


def save_config(config):
    if not os.path.exists(CONFIG_PATH):
        os.makedirs(CONFIG_PATH)
    path = CONFIG_PATH + CONFIG_FILE
    with open(path, 'wb') as f:
        json.dump(config, f)


def start_log(level=logging.INFO):
    observer = logger(level)
    log.startLoggingWithObserver(observer)


def logger(loglevel):
    numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % loglevel)
    logging.basicConfig(
        format='%(asctime)s-[%(levelname)s]-%(name)s :  %(message)s',
        level=numeric_level)
    observer = log.PythonLoggingObserver()
    return observer.emit



#!/usr/bin/env python

# Copyright (c) 2014 clowwindy
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import sys
import hashlib
import string
import struct
import logging
import encrypt_salsa20


def random_string(length):
    try:
        import M2Crypto.Rand
        return M2Crypto.Rand.rand_bytes(length)
    except ImportError:
        # TODO really strong enough on Linux?
        return os.urandom(length)


cached_tables = {}
cached_keys = {}


def get_table(key):
    m = hashlib.md5()
    m.update(key)
    s = m.digest()
    (a, b) = struct.unpack('<QQ', s)
    table = [c for c in string.maketrans('', '')]
    for i in xrange(1, 1024):
        table.sort(lambda x, y: int(a % (ord(x) + i) - a % (ord(y) + i)))
    return table


def init_table(key, method=None):
    if method is not None and method == 'table':
        method = None
    if method:
        try:
            __import__('M2Crypto')
        except ImportError:
            logging.error(('M2Crypto is required to use %s, please run'
                           ' `apt-get install python-m2crypto`') % method)
            sys.exit(1)
    if not method:
        if key in cached_tables:
            return cached_tables[key]
        encrypt_table = ''.join(get_table(key))
        decrypt_table = string.maketrans(encrypt_table,
                                         string.maketrans('', ''))
        cached_tables[key] = [encrypt_table, decrypt_table]
    else:
        try:
            Encryptor(key, method)  # test if the settings if OK
        except Exception as e:
            logging.error(e)
            sys.exit(1)


def EVP_BytesToKey(password, key_len, iv_len):
    # equivalent to OpenSSL's EVP_BytesToKey() with count 1
    # so that we make the same key and iv as nodejs version
    password = str(password)
    r = cached_keys.get(password, None)
    if r:
        return r
    m = []
    i = 0
    while len(''.join(m)) < (key_len + iv_len):
        md5 = hashlib.md5()
        data = password
        if i > 0:
            data = m[i - 1] + password
        md5.update(data)
        m.append(md5.digest())
        i += 1
    ms = ''.join(m)
    key = ms[:key_len]
    iv = ms[key_len:key_len + iv_len]
    cached_keys[password] = (key, iv)
    return (key, iv)


method_supported = {
    'aes-128-cfb': (16, 16),
    'aes-192-cfb': (24, 16),
    'aes-256-cfb': (32, 16),
    'bf-cfb': (16, 8),
    'camellia-128-cfb': (16, 16),
    'camellia-192-cfb': (24, 16),
    'camellia-256-cfb': (32, 16),
    'cast5-cfb': (16, 8),
    'des-cfb': (8, 8),
    'idea-cfb': (16, 8),
    'rc2-cfb': (16, 8),
    'rc4': (16, 0),
    'seed-cfb': (16, 16),
    'salsa20-ctr': (32, 8),
}


class Encryptor(object):
    def __init__(self, key, method=None):
        if method == 'table':
            method = None
        self.key = key
        self.method = method
        self.iv = None
        self.iv_sent = False
        self.cipher_iv = ''
        self.decipher = None
        if method:
            self.cipher = self.get_cipher(key, method, 1, iv=random_string(32))
        else:
            self.encrypt_table, self.decrypt_table = init_table(key)
            self.cipher = None

    def get_cipher_len(self, method):
        method = method.lower()
        m = method_supported.get(method, None)
        return m

    def iv_len(self):
        return len(self.cipher_iv)

    def get_cipher(self, password, method, op, iv=None):
        password = password.encode('utf-8')
        method = method.lower()
        m = self.get_cipher_len(method)
        if m:
            key, iv_ = EVP_BytesToKey(password, m[0], m[1])
            if iv is None:
                iv = iv_
            iv = iv[:m[1]]
            if op == 1:
                self.cipher_iv = iv[:m[1]]  # this iv is for cipher not decipher
            if method != 'salsa20-ctr':
                import M2Crypto.EVP
                return M2Crypto.EVP.Cipher(method.replace('-', '_'), key, iv,
                                           op, key_as_bytes=0, d='md5',
                                           salt=None, i=1, padding=1)
            else:
                return encrypt_salsa20.Salsa20Cipher(method, key, iv, op)

        logging.error('method %s not supported' % method)
        sys.exit(1)

    def encrypt(self, buf):
        if len(buf) == 0:
            return buf
        if not self.method:
            return string.translate(buf, self.encrypt_table)
        else:
            if self.iv_sent:
                return self.cipher.update(buf)
            else:
                self.iv_sent = True
                return self.cipher_iv + self.cipher.update(buf)

    def decrypt(self, buf):
        if len(buf) == 0:
            return buf
        if not self.method:
            return string.translate(buf, self.decrypt_table)
        else:
            if self.decipher is None:
                decipher_iv_len = self.get_cipher_len(self.method)[1]
                decipher_iv = buf[:decipher_iv_len]
                self.decipher = self.get_cipher(self.key, self.method, 0,
                                                iv=decipher_iv)
                buf = buf[decipher_iv_len:]
                if len(buf) == 0:
                    return buf
            return self.decipher.update(buf)


def encrypt_all(password, method, op, data):
    if method is not None and method.lower() == 'table':
        method = None
    if not method:
        [encrypt_table, decrypt_table] = init_table(password)
        if op:
            return string.translate(data, encrypt_table)
        else:
            return string.translate(data, decrypt_table)
    else:
        import M2Crypto.EVP
        result = []
        method = method.lower()
        (key_len, iv_len) = method_supported[method]
        (key, _) = EVP_BytesToKey(password, key_len, iv_len)
        if op:
            iv = random_string(iv_len)
            result.append(iv)
        else:
            iv = data[:iv_len]
            data = data[iv_len:]
        if method != 'salsa20-ctr':
            cipher = M2Crypto.EVP.Cipher(method.replace('-', '_'), key, iv,
                                         op, key_as_bytes=0, d='md5',
                                         salt=None, i=1, padding=1)
        else:
            cipher = encrypt_salsa20.Salsa20Cipher(method, key, iv, op)
        result.append(cipher.update(data))
        return ''.join(result)



from setuptools import setup, find_packages

version = '0.1'

setup(
    name='shadowsocks-gtk',
    version=version,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=True,
    author='apporc',
    author_email='apporc@gmail.com',
    url='https://github.com/apporc/shadowsocks-gtk',
    license='http://opensource.org/licenses/MIT',
    description='ShadowSocks Gtk Client',
    install_requires=[],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: X11 Applications',
        'Environment :: X11 Applications :: Gnome',
        'Environment :: X11 Applications :: GTK',
        'Framework :: Twisted',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: Chinese (Simplified)',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet',
        ],
    entry_points={
        'console_scripts': [
            'sslocal = shadowsocks_gtk.local:main',
        ],
        'gui_scripts': [
            'shadowsocks-gtk = shadowsocks_gtk.shadowsocks:main',
        ]
    }
)



# Copyright (c) 2014 apporc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from twisted.internet import protocol, reactor
from twisted.protocols.policies import TimeoutMixin
import struct
from utils import get_config, start_log
import logging
logger = logging.getLogger()

import encrypt
config = get_config()


STATE_INITIAL = 0
STATE_AUTH = 1
STATE_REQUEST = 2
STATE_READY = 3
STATE_AUTH_USERPASS = 4
STATE_LAST = 5

STATE_CONNECT_PENDING = STATE_LAST + 1

SOCKS5_VER = 0x05

ADDR_IPV4 = 0x01
ADDR_DOMAINNAME = 0x03
ADDR_IPV6 = 0x04

CMD_CONNECT = 0x01
CMD_BIND = 0x02
CMD_UDPASSOC = 0x03

AUTHMECH_ANON = 0x00
AUTHMECH_USERPASS = 0x02
AUTHMECH_INVALID = 0xFF

REPLY_SUCCESS = 0x00
REPLY_GENERAL_FAILUR = 0x01
REPLY_CONN_NOT_ALLOWED = 0x02
REPLY_NETWORK_UNREACHABLE = 0x03
REPLY_HOST_UNREACHABLE = 0x04
REPLY_CONN_REFUSED = 0x05
REPLY_TTL_EXPIRED = 0x06
REPLY_CMD_NOT_SUPPORTED = 0x07
REPLY_ADDR_NOT_SUPPORTED = 0x08


class SOCKSv5Outgoing(protocol.Protocol):
    def __init__(self, peersock):
        self.peersock = peersock
        self.peersock.peersock = self

    def connectionMade(self):
        logger.info('Connected to %s' % self.peersock.server)
        self.peersock.connectCompleted()

    def connectionLost(self, reason):
        logger.info('Lost connection to %s' % self.peersock.server)
        self.peersock.transport.loseConnection()
        self.peersock.peersock = None
        self.peersock = None

    def dataReceived(self, buf):
        buf = self.peersock.encryptor.decrypt(buf)
        self.peersock.transport.write(buf)


class SOCKSv5(protocol.Protocol, TimeoutMixin):

    inputTimeout = 600

    def __init__(self):
        self.server = config['server']
        self.server_port = config['server_port']
        self.data_to_send = []
        self.state = STATE_INITIAL
        self.buf = ""
        self.supportedAuthMechs = [AUTHMECH_ANON, AUTHMECH_USERPASS]
        self.supportedAddrs = [ADDR_IPV4, ADDR_DOMAINNAME]
        self.enabledCommands = [CMD_CONNECT, CMD_UDPASSOC]
        self.peersock = None
        self.addressType = 0
        self.requestType = 0
        self.encryptor = encrypt.Encryptor(config['password'],
                                           config['method'])

    def _parseNegotiation(self):
        try:
            # Parse out data
            ver, nmethod = struct.unpack('!BB', self.buf[:2])
            methods = struct.unpack('%dB' % nmethod, self.buf[2:nmethod+2])

            # Ensure version is correct
            if ver != 5:
                self.transport.write(struct.pack('!BB', SOCKS5_VER, AUTHMECH_INVALID))
                self.transport.loseConnection()
                return

            # Trim off front of the buffer
            self.buf = self.buf[nmethod+2:]

            # Check for supported auth mechs
            for m in self.supportedAuthMechs:
                if m in methods:
                    # Update internal state, according to selected method
                    if m == AUTHMECH_ANON:
                        self.state = STATE_REQUEST
                    elif m == AUTHMECH_USERPASS:
                        self.state = STATE_AUTH_USERPASS
                    # Complete negotiation w/ this method
                    self.transport.write(struct.pack('!BB', SOCKS5_VER, m))
                    return

            # No supported mechs found, notify client and close the connection
            logger.error('Not supported mechs found')
            self.transport.write(struct.pack('!BB', SOCKS5_VER, AUTHMECH_INVALID))
            self.transport.loseConnection()
        except struct.error:
            pass

    def _parseUserPass(self):
        try:
            # Parse out data
            ver, ulen = struct.unpack('BB', self.buf[:2])
            uname, = struct.unpack('%ds' % ulen, self.buf[2:ulen + 2])
            plen, = struct.unpack('B', self.buf[ulen + 2])
            password, = struct.unpack('%ds' % plen, self.buf[ulen + 3:ulen + 3 + plen])
            # Trim off fron of the buffer
            self.buf = self.buf[3 + ulen + plen:]
            # Fire event to authenticate user
            if self.authenticateUserPass(uname, password):
                # Signal success
                self.state = STATE_REQUEST
                self.transport.write(struct.pack('!BB', SOCKS5_VER, 0x00))
            else:
                # Signal failure
                self.transport.write(struct.pack('!BB', SOCKS5_VER, 0x01))
                self.transport.loseConnection()
        except struct.error:
            pass

    def sendErrorReply(self, errorcode):
        # Any other address types are not supported
        result = struct.pack('!BBBBIH', SOCKS5_VER, errorcode, 0, 1, 0, 0)
        self.transport.write(result)
        self.transport.loseConnection()

    def _parseRequest(self):
        try:
            # Parse out data and trim buffer accordingly
            ver, cmd, rsvd, self.addressType = struct.unpack('!BBBB', self.buf[:4])

            # Ensure command is supported
            if cmd not in self.enabledCommands:
                # Send a not supported error
                self.sendErrorReply(REPLY_CMD_NOT_SUPPORTED)
                return

            # Process the command
            if cmd == CMD_CONNECT:
                result = '\x05\x00\x00\x01\x00\x00\x00\x00\x10\x10'
                self.transport.write(result)
                self.data_to_send.append(self.encryptor.encrypt(self.buf[3:]))
                self.connectServer(self.server, self.server_port)
            elif cmd == CMD_UDPASSOC:
                logger .warn('Got UDP ASSOC request!!')
            else:
                # Any other command is not supported
                self.sendErrorReply(REPLY_CMD_NOT_SUPPORTED)

        except struct.error:
            return None

    def connectServer(self, addr, port):
        self.transport.stopReading()
        self.state = STATE_CONNECT_PENDING
        protocol.ClientCreator(reactor, SOCKSv5Outgoing, self).connectTCP(addr, port)

    def connectCompleted(self):
        self.state = STATE_READY
        self.transport.startReading()

    def authenticateUserPass(self, user, passwd):
        logger.info("User/pass: %s/%s" % (user, passwd))
        return True

    def connectionMade(self):
        self.setTimeout(self.inputTimeout)

    def dataReceived(self, buf):
        self.resetTimeout()

        if self.state == STATE_READY:
            self.data_to_send.append(self.encryptor.encrypt(buf))
            data = ''.join(self.data_to_send)
            self.data_to_send = []
            self.peersock.transport.write(data)
            return

        self.buf = self.buf + buf
        if self.state == STATE_INITIAL:
            self._parseNegotiation()
        if self.state == STATE_AUTH_USERPASS:
            self._parseUserPass()
        if self.state == STATE_REQUEST:
            self._parseRequest()

    def timeoutConnection(self):
        logger.warn('Connection timedout!')
        self.transport.unregisterProducer()
        self.transport.loseConnection()


def main():
    factory = protocol.Factory()
    factory.protocol = SOCKSv5
    start_log(config['level'])
    encrypt.init_table(config['password'], config['method'])
    reactor.listenTCP(config['local_port'], factory)
    reactor.run()

if __name__ == "__main__":
    main()
